from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional

from .prediction_timing import PredictionTiming

class Prediction(BaseModel):
    id: int = Field(alias='id')
    operation_type: int = Field(alias='operationType')
    vehicle_id: str = Field(alias='vehicleId')
    naptan_id: str = Field(alias='naptanId')
    station_name: str = Field(alias='stationName')
    line_id: str = Field(alias='lineId')
    line_name: str = Field(alias='lineName')
    platform_name: Optional[str] = Field(None, alias='platformName')
    direction: Optional[str] = Field(None, alias='direction')
    bearing: str = Field(alias='bearing')
    destination_naptan_id: str = Field(alias='destinationNaptanId')
    destination_name: str = Field(alias='destinationName')
    timestamp: datetime = Field(alias='timestamp')
    time_to_station: int = Field(alias='timeToStation')
    current_location: str = Field(alias='currentLocation')
    towards: str = Field(alias='towards')
    expected_arrival: datetime = Field(alias='expectedArrival')
    time_to_live: datetime = Field(alias='timeToLive')
    mode_name: str = Field(alias='modeName')
    timing: PredictionTiming = Field(alias='timing')
    content_expires: Optional[datetime] = Field(None)
    shared_expires: Optional[datetime] = Field(None)

    model_config = {'populate_by_name': True}


from pydantic import BaseModel, Field

from .stop_point import StopPoint

class RouteSectionNaptanEntrySequence(BaseModel):
    ordinal: int = Field(alias='ordinal')
    stop_point: StopPoint = Field(alias='stopPoint')

    model_config = {'populate_by_name': True}



# from https://github.com/dhilmathy/TfL-python-api
# MIT License

# Copyright (c) 2018 Mathivanan Palanisamy

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

base_url = "https://api.tfl.gov.uk/"

endpoints = {
    'stopPointsByLineId': {"uri": 'Line/{0}/StopPoints', "model" : "StopPoint"},
    'lineMetaModes': {"uri": 'Line/Meta/Modes', "model" : "Mode"},
    'linesByLineId': {"uri": 'Line/{0}', "model" : "Line"},
    'linesByMode': {"uri": 'Line/Mode/{0}', "model" : "Line"},
    'lineStatus': {"uri": 'Line/{0}/Status', "model" : "Line"},
    'lineStatusBySeverity': {"uri": 'Line/Status/{0}', "model" : "Line"},
    'lineStatusByMode': {"uri": 'Line/Mode/{0}/Status', "model" : "Line"},
    'routeByLineId': {"uri": 'Line/{0}/Route', "model" : "Line"},
    'routeByLineIdWithDirection': {"uri": 'Line/{0}/Route/Sequence/{1}', "model" : "RouteSequence"},
    'routeByMode': {"uri": 'Line/Mode/{0}/Route', "model" : "Line"},
    'lineDisruptionsByLineId': {"uri": 'Line/{0}/Disruption', "model" : "Disruption"},
    'lineDisruptionsByMode': {"uri": 'Line/Mode/{0}/Disruption', "model" : "Disruption"},
    'stopPointMetaModes': {"uri": 'StopPoint/Meta/Modes', "model" : "Mode"},
    'stopPointById': {"uri":  'StopPoint/{0}', "model" : "StopPoint"},
    'stopPointByMode':  {"uri": 'StopPoint/Mode/{0}', "model" : "StopPointsResponse"},
    'arrivalsByLineId': {"uri": 'Line/{0}/Arrivals', "model" : "Prediction"}
}



from pydantic import BaseModel, Field
from typing import List, Optional
from datetime import datetime


from .line_group import LineGroup
from .line_mode_group import LineModeGroup
from .additional_properties import AdditionalProperties
# note additional import from .line import Line
# at the end of the file

class StopPoint(BaseModel):
    naptan_id: str = Field(alias="naptanId")
    platform_name: Optional[str] = Field(None, alias="platformName")
    indicator: Optional[str] = Field(None, alias="indicator")
    stop_letter: Optional[str] = Field(None, alias="stopLetter")
    modes: List[str] = Field(alias="modes")
    ics_code: Optional[str] = Field(None, alias="icsCode")
    sms_code: Optional[str] = Field(None, alias="smsCode")
    stop_type: Optional[str] = Field(None, alias="stopType")
    station_naptan: Optional[str] = Field(None, alias="stationNaptan")
    accessibility_summary: Optional[str] = Field(None, alias="accessibilitySummary")
    hub_naptan_code: Optional[str] = Field(None, alias="hubNaptanCode")
    lines: List["Line"] = Field(alias="lines")
    line_group: List[LineGroup] = Field(alias="lineGroup")
    line_mode_groups: List[LineModeGroup] = Field(alias="lineModeGroups")
    full_name: Optional[str] = Field(None, alias="fullName")
    naptan_mode: Optional[str] = Field(None, alias="naptanMode")
    status: Optional[bool] = Field(None, alias="status")
    id: str = Field(alias="id")
    url: Optional[str] = Field(None, alias="url")
    common_name: str = Field(alias="commonName")
    distance: Optional[int] = Field(None, alias="distance")
    place_type: str = Field(alias="placeType")
    additional_properties: List[AdditionalProperties] = Field(
        alias="additionalProperties"
    )
    children: Optional[List["StopPoint"]] = Field(None, alias="children")
    children_urls: Optional[List[str]] = Field([], alias="childrenUrls")
    lat: float = Field(alias="lat")
    lon: float = Field(alias="lon")
    content_expires: Optional[datetime] = Field(None)
    shared_expires: Optional[datetime] = Field(None)

    model_config = {"populate_by_name": True}


# we have to import Line after StopPoint is defined or we get a circular reference
from .line import Line  # noqa: E402

# Rebuild the model to reference Line
StopPoint.model_rebuild()



from pydantic import BaseModel, Field
from typing import List
from datetime import datetime

from .route_section_naptan_entry_sequence import RouteSectionNaptanEntrySequence

class AffectedRoute(BaseModel):
    id: str = Field(alias='id')
    line_id: str = Field(alias='lineId')
    route_code: str = Field(alias='routeCode')
    name: str = Field(alias='name')
    line_string: str = Field(alias='lineString')
    direction: str = Field(alias='direction')
    origination_name: str = Field(alias='originationName')
    destination_name: str = Field(alias='destinationName')
    valid_to: datetime = Field(alias='validTo')
    valid_from: datetime = Field(alias='validFrom')
    route_section_naptan_entry_sequence: List[RouteSectionNaptanEntrySequence] = Field(alias='routeSectionNaptanEntrySequence')

    model_config = {'populate_by_name': True}



from pydantic import BaseModel, Field
from datetime import datetime
from typing import List, Optional

from .affected_route import AffectedRoute
from .stop_point import StopPoint


class Disruption(BaseModel):
    category: str = Field(alias='category')
    disruption_type: str = Field(alias='type')
    category_description: str = Field(alias='categoryDescription')
    description: str = Field(alias='description')
    summary: Optional[str] = Field(None, alias='summary')
    additional_info: Optional[str] = Field(None, alias='additionalInfo')
    created: Optional[datetime] = Field(None, alias='created')
    last_update: Optional[datetime] = Field(None, alias='lastUpdate')
    affected_routes: List[AffectedRoute] = Field(alias='affectedRoutes')
    affected_stops: List[StopPoint] = Field(alias='affectedStops')
    closure_text: str = Field(alias='closureText')
    content_expires: Optional[datetime] = Field(None)
    shared_expires: Optional[datetime] = Field(None)

    model_config = {'populate_by_name': True}

    


# portions of this code are from https://github.com/dhilmathy/TfL-python-api
# MIT License

# Copyright (c) 2018 Mathivanan Palanisamy
# Copyright (c) 2024 Rob Aleck

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .config import endpoints
from .rest_client import RestClient
from importlib import import_module
from typing import Any, Literal, List, Optional, Tuple
from requests import Response
import pkgutil
from pydantic import BaseModel
from . import models
from datetime import datetime, timedelta
from email.utils import parsedate_to_datetime


class Client:
    """Client

    :param str api_token: API token to access TfL unified API
    """

    def __init__(self, api_token: str = None):
        self.client = RestClient(api_token)
        self.models = self._load_models()

    def _load_models(self):
        models_dict = {}
        for importer, modname, ispkg in pkgutil.iter_modules(models.__path__):
            module = import_module(f".models.{modname}", __package__)
            for model_name in dir(module):
                attr = getattr(module, model_name)
                if isinstance(attr, type) and issubclass(attr, BaseModel):
                    models_dict[model_name] = attr
        # print(models_dict)
        return models_dict

    @staticmethod
    def _parse_int_or_none(value: str) -> int | None:
        try:
            return int(value)
        except (TypeError, ValueError):
            return None

    @staticmethod
    def _get_maxage_headers_from_cache_control_header(response: Response) -> Tuple[Optional[int], Optional[int]]:
        cache_control = response.headers.get("Cache-Control")
        # e.g. 'public, must-revalidate, max-age=43200, s-maxage=86400'
        if cache_control is None:
            return None, None
        directives = cache_control.split(", ")
        # e.g. ['public', 'must-revalidate', 'max-age=43200', 's-maxage=86400']
        directives = {d.split("=")[0]: d.split("=")[1]
                      for d in directives if "=" in d}
        smaxage = Client._parse_int_or_none(directives.get("s-maxage", ""))
        maxage = Client._parse_int_or_none(directives.get("max-age", ""))
        return smaxage, maxage

    @staticmethod
    def _parse_timedelta(value: Optional[int], base_time: Optional[datetime]) -> Optional[datetime]:
        try:
            return base_time + timedelta(seconds=value) if value is not None and base_time is not None else None
        except (TypeError, ValueError):
            return None

    @staticmethod
    def _get_result_expiry(response: Response) -> Tuple[datetime | None, datetime | None]:
        s_maxage, maxage = Client._get_maxage_headers_from_cache_control_header(
            response)
        request_datetime = parsedate_to_datetime(response.headers.get(
            "Date")) if "Date" in response.headers else None

        s_maxage_expiry = Client._parse_timedelta(s_maxage, request_datetime)
        maxage_expiry = Client._parse_timedelta(maxage, request_datetime)

        return s_maxage_expiry, maxage_expiry

    def _deserialize(self, model_name: str, response: Response) -> Any:
        shared_expiry, result_expiry = self._get_result_expiry(response)
        Model = self._get_model(model_name)
        data = response.json()

        result = self._create_model_instance(
            Model, data, result_expiry, shared_expiry)

        return result

    def _get_model(self, model_name: str) -> BaseModel:
        Model = self.models.get(model_name)
        if Model is None:
            raise ValueError(f"No model found with name {model_name}")
        return Model

    def _create_model_instance(
        self, Model: BaseModel,
        response_json: Any,
        result_expiry: datetime | None,
        shared_expiry: datetime | None
    ) -> BaseModel | List[BaseModel]:
        if isinstance(response_json, dict):
            return self._create_model_with_expiry(Model, response_json, result_expiry, shared_expiry)
        else:
            return [
                self._create_model_with_expiry(
                    Model, item, result_expiry, shared_expiry)
                for item in response_json
            ]

    def _create_model_with_expiry(
        self, Model: BaseModel, response_json: Any, result_expiry: Optional[datetime], shared_expiry: Optional[datetime]
    ):
        instance = Model(**response_json)
        instance.content_expires = result_expiry
        instance.shared_expires = shared_expiry
        return instance

    def _deserialize_error(self, response: Response) -> models.ApiError:
        # if content is json, deserialize it, otherwise manually create an ApiError object
        if response.headers.get("Content-Type") == "application/json":
            return self._deserialize("ApiError", response)
        return models.ApiError(
            timestampUtc=parsedate_to_datetime(response.headers.get("Date")),
            exceptionType="Unknown",
            httpStatusCode=response.status_code,
            httpStatus=response.reason,
            relativeUri=response.url,
            message=response.text,
        )

    def _send_request_and_deserialize(
        self, endpoint_and_model: dict[str, str],
        params: str | int | List[str | int] = None, endpoint_args: dict = None
    ) -> BaseModel | List[BaseModel] | models.ApiError:
        if params is None:
            params = []
        if not isinstance(params, list):
            params = [params]

        endpoint = endpoint_and_model["uri"].format(*params)
        model_name = endpoint_and_model["model"]

        response = self.client.send_request(endpoint, endpoint_args)

        if response.status_code != 200:
            return self._deserialize_error(response)
        return self._deserialize(model_name, response)

    def get_stop_points_by_line_id(
        self, line_id: str
    ) -> models.StopPoint | List[models.StopPoint] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["stopPointsByLineId"], line_id)

    def get_line_meta_modes(self) -> models.Mode | models.ApiError:
        return self._send_request_and_deserialize(endpoints["lineMetaModes"])

    def get_lines(
        self, line_id: str | None = None, mode: str | None = None
    ) -> models.Line | List[models.Line] | models.ApiError:
        if line_id is None and mode is None:
            raise ValueError(
                "Either the --line_id argument or the --mode argument needs to be specified."
            )
        if line_id is not None:
            return self._send_request_and_deserialize(endpoints["linesByLineId"], line_id)
        return self._send_request_and_deserialize(endpoints["linesByMode"], mode)

    def get_line_status(
        self, line: str, include_details: bool = None
    ) -> models.Line | List[models.Line] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["lineStatus"], line, {"detail": include_details}
        )

    def get_line_status_severity(
        self, severity: str
    ) -> models.Line | List[models.Line] | models.ApiError:
        """
        severity: The level of severity (eg: a number from 0 to 14)
        """
        return self._send_request_and_deserialize(
            endpoints["lineStatusBySeverity"], severity
        )

    def get_line_status_by_mode(
        self, mode: str
    ) -> models.Line | List[models.Line] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["lineStatusByMode"], mode
        )

    def get_route_by_line_id(
        self, line_id: str
    ) -> models.Line | List[models.Line] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["routeByLineId"], line_id
        )

    def get_route_by_mode(
        self, mode: str, service_types: Optional[List[Literal["regular", "night"]]] = None
    ) -> models.Line | List[models.Line] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["routeByMode"], mode, {"serviceTypes": service_types}
        )

    def get_route_by_line_id_with_direction(
        self, line_id: str, direction: Literal["inbound", "outbound", "all"]
    ) -> models.RouteSequence | List[models.RouteSequence] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["routeByLineIdWithDirection"], [line_id, direction]
        )

    def get_line_disruptions_by_line_id(
        self, line_id: str
    ) -> models.Disruption | List[models.Disruption] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["lineDisruptionsByLineId"], line_id
        )

    def get_line_disruptions_by_mode(
        self, mode: str
    ) -> models.Disruption | List[models.Disruption] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["lineDisruptionsByMode"], mode
        )

    def get_stop_points_by_id(
        self, id: str
    ) -> models.StopPoint | List[models.StopPoint] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["stopPointById"], id
        )

    def get_stop_points_by_mode(
        self, mode: str
    ) -> models.StopPointsResponse | List[models.StopPointsResponse] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["stopPointByMode"], mode
        )

    def get_stop_point_meta_modes(
        self,
    ) -> models.Mode | List[models.Mode] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["stopPointMetaModes"]
        )

    def get_arrivals_by_line_id(
        self, line_id: str
    ) -> models.Prediction | List[models.Prediction] | models.ApiError:
        return self._send_request_and_deserialize(
            endpoints["arrivalsByLineId"], line_id
        )



from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional

class Mode(BaseModel):
    is_tfl_service: bool = Field(alias='isTflService')
    is_fare_paying: bool = Field(alias='isFarePaying')
    is_scheduled_service: bool = Field(alias='isScheduledService')
    mode_name: str = Field(alias='modeName')
    content_expires: Optional[datetime] = Field(None)
    shared_expires: Optional[datetime] = Field(None)

    model_config = {'populate_by_name': True}


from pydantic import BaseModel, Field
from typing import List, Optional

class LineGroup(BaseModel):
    naptan_id_reference: Optional[str] = Field(None, alias='naptanIdReference')
    station_atco_code: Optional[str] = Field(None, alias='stationActoCode')
    line_identifier: Optional[List[str]] = Field([], alias='lineIdentifier')

    model_config = {'populate_by_name': True}


from pydantic import BaseModel, Field
from typing import Optional

from .identifier import Identifier

class MatchedStop(BaseModel):
    station_id: Optional[str] = Field(None, alias='stationId')
    ics_id: str = Field(alias='icsId')
    top_most_parent_id: Optional[str] = Field(None, alias='topMostParentId')
    modes: list[str]
    stop_type: str = Field(alias='stopType')
    zone: Optional[str] = Field(None, alias='zone')
    lines: Optional[list[Identifier]] = Field([], alias='lines')
    status: bool = Field(alias='status')
    id: str = Field(alias='id')
    name: str = Field(alias='name')
    lat: float = Field(alias='lat')
    lon: float = Field(alias='lon')

    model_config = {'populate_by_name': True}



from pydantic import BaseModel, Field
from datetime import datetime
from typing import List, Optional

from .validity_period import ValidityPeriod
from .disruption import Disruption

class LineStatus(BaseModel):
    id: int = Field(alias='id')
    line_id: Optional[str] = Field(None, alias='lineId')
    status_severity: int = Field(alias='statusSeverity')
    status_severity_description: str = Field(alias='statusSeverityDescription')
    reason: Optional[str] = Field(None, alias='reason')
    created: datetime = Field(alias='created')
    last_update: Optional[datetime] = Field(None, alias='lastUpdate')
    validity_periods: List[ValidityPeriod] = Field(None, alias='validityPeriods')
    disruptions: Optional[List[Disruption]] = Field([], alias='disruptions')

    model_config = {'populate_by_name': True}